# using Click implementation
import datetime
import os
import re
import subprocess
from functools import wraps

from pathlib import Path
from typing import List, Optional, Tuple

import click  # CLI magic
from git import Repo

from todotree.Errors.TodoFileNotFound import TodoFileNotFound
from todotree.Errors.DoneFileNotFound import DoneFileNotFound
from todotree.Errors.ConfigFileNotFound import ConfigFileNotFound
from todotree.Taskmanager import Taskmanager, task_to_done
from todotree.config import Config


# Click Replaces:
# Argparse
#
# NOTE: this file cannot be in a class. See: https://github.com/pallets/click/issues/601
# But context and variable ferrying can be done using the context option.
# We just call the context 'self' and hope the issue does resolve itself.
# https://click.palletsprojects.com/en/8.1.x/commands/#nested-handling-and-contexts

def common_options(function):
    """
    Wrapper that defines common functions.

    It should be used as a decorator: `@common_options`

    This function will be needed when we want to support both
    `todotree cd --verbose` and `todotree --verbose cd`
    """

    @wraps(function)
    @click.option('--config-path', default=None, help="Path to the configuration file.")
    @click.option('--todo-file', default=None, help="Path to the todo.txt file, overrides --config.")
    @click.option('--done-file', default=None, help="Path to the done.txt file, overrides --config.")
    @click.option('--verbose', is_flag=True, help="Increases verbosity in messages.", is_eager=True)
    @click.option('--quiet', is_flag=True, help="Do not print messages, only output. Useful in scripts.", is_eager=True)
    def wrapper(*args, **kwargs):
        return function(*args, **kwargs)

    return wrapper


@click.group()
@common_options
@click.pass_context
def root(self: click.Context, config_path: Optional[Path], todo_file: Optional[Path], done_file: Optional[Path],
         verbose: bool, quiet: bool):
    """todotree help text."""
    # ^ This text also shows up in the help command.
    # Root click group. This manages all the command line interactions.
    # ensure that ctx.obj exists and is a dict
    self.ensure_object(dict)
    initialize(config_path, done_file, quiet, self, todo_file, verbose)
    # Pass to another command.
    pass


def initialize(config_path: Optional[Path], done_file: Optional[Path], quiet: bool, self: click.Context,
               todo_file: Optional[Path], verbose: bool):
    """
    Initializes the application and loads the configuration.
    :param config_path: Path to the configuration file from the command line.
    :param done_file: Path to the configuration file from the command line.
    :param quiet: --quiet value
    :param self: Context.
    :param todo_file: Path to the configuration file from the command line.
    :param verbose: --verbose value.
    """
    # parsing arguments.
    config = Config()
    try:
        config.read(config_path)
    except ConfigFileNotFound as e:
        handle_config_file_not_found(e, verbose, config_path)

    if todo_file is not None:
        config.todo_file = todo_file
    if done_file is not None:
        config.done_file = done_file
    if verbose:
        config.verbose = True
    if quiet:
        config.quiet = True
        config.verbose = False
    # Logging
    if config.verbose:
        click.echo(config.consoleGood + f"Read configuration from {config.config_file}")
        click.echo(config.consoleGood + f"The todo file is supposed to be at {config.todo_file}")
        click.echo(config.consoleGood + f"The done file is supposed to be at {config.done_file}")
    # creating variables a la __init__.
    self.obj["config"] = config
    self.obj["task_manager"] = Taskmanager(configuration=config)


def handle_config_file_not_found(e: BaseException, verbose: bool, config_path):
    """Handle when the configuration file is not found.

    :param config_path: The path where the file was not found.
    :param verbose: Additional verbosity.
    :param e: The exception raised.
    """
    click.echo(Config().consoleWarn + " The config.yaml file could not be found.")
    if verbose:
        click.echo(f"The config file should be located at {config_path}")
        click.echo(e)
    click.echo("The default options are now used.")


def handle_todo_file_not_found(e: Exception, self):
    """Inform the user that the todo.txt was not found."""
    click.echo(f"{self.obj['config'].consoleError} The todo.txt could not be found.")
    click.echo(
        f"{self.obj['config'].consoleError} It searched at the following location: {self.obj['config'].todo_file}")
    if self.obj["config"].verbose:
        click.echo(e)


def commit_exit(action: str, config: Config):
    """
    Commit the files with git before exiting.

    :param config: The configuration parameters.
    :param action: The name of the action, such as list or add.
    """
    if config.git_mode != "Local":
        exit()
    repo = Repo(config.todo_folder)

    # Git add.
    repo.index.add('*')

    # Git commit.
    time = datetime.datetime.now().isoformat()
    repo.index.commit(message=time + action)

    if config.git_mode == "Full":
        # Git push.
        repo.remote().push()


@root.command('add', short_help='Add a task to the task list')
@click.argument('task', type=str, nargs=-1)  # , short_help = 'Task To Add')
@click.pass_context
def add(self, task: Tuple):
    # Convert tuple to string
    task: str = " ".join(map(str, task))
    try:
        # Disable fancy imports, because they are not needed.
        self.obj['config'].enable_project_folder = False
        # Import tasks.
        self.obj["task_manager"].import_tasks()
        # Add task
        new_number = self.obj["task_manager"].add_task_to_file(task.strip() + "\n")
        click.echo(f"{self.obj['config'].consoleGood} Task added:")
        click.echo(f"{new_number} {task}")
        commit_exit("add", self.obj["config"])
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)


@root.command('addx', short_help='Add a task and immediately mark it as done')
@click.argument('task', type=str)
@click.pass_context
def add_x(self, task):
    try:
        f = open(self.obj["config"].done_file, "a")
        done = task_to_done(
            task.strip())
        f.write(done)
        click.echo(done)
    except FileNotFoundError as e:
        handle_done_file_not_found(e, self)
        exit(1)


def handle_done_file_not_found(e, self):
    """Inform the user that the done.txt was not found."""
    click.echo(self.obj["config"].consoleError + " The done.txt could not be found.")
    click.echo(
        f"{self.obj['config'].consoleError} It searched at the following location: {self.obj['config'].done_file}")
    if self.obj["config"].verbose:
        click.echo(e)


@root.command('append', short_help='append `append_string` to `task_nr`')
@click.argument('task_nr', type=int)
@click.argument('append_string', type=str, nargs=-1)
@click.pass_context
def append(self, task_nr: int, append_string: str):
    # Disable fancy imports, because they are not needed.
    self.obj['config'].enable_project_folder = False
    # Import tasks.
    try:
        self.obj["task_manager"].import_tasks()
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)
    # Convert tuple to string.
    append_string = " ".join(append_string)
    click.echo(f"{self.obj['config'].consoleGood} The new task is: ")
    click.echo(self.obj["task_manager"].append_to_task(task_nr, append_string.strip()))
    commit_exit("append", self.obj["config"])


@root.command('context',
              short_help='list task in a tree, by context',
              help='list a tree, of which the first node is the context, the second nodes are the tasks')
@click.pass_context
def context(self):
    try:
        self.obj['task_manager'].import_tasks()
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)
    # Print due tree.
    click.echo(self.obj['task_manager'].print_context_tree())
    exit()


@root.command('cd',
              short_help='print directory of the todo.txt directory',
              help='print directory of the todo.txt directory'
              )
@click.pass_context
def cd(self):
    config_path: Path = Path(self.obj['config'].todo_folder)
    if self.obj['config'].verbose:
        click.echo(f"{self.obj['config'].consoleGood} The location to the data folder is: ")

    if config_path.is_absolute():
        # Then the configured path is printed.
        click.echo(str(config_path))
        exit()
    # Then the relative path is resolved to be absolute.
    base_path: Path = Path.home()
    full_path: Path = base_path / config_path
    click.echo(str(full_path))
    exit()


@root.command('do',
              short_help='mark task as done and move it to the done.txt'
              )
@click.argument('task_numbers', type=list, nargs=-1)  # type=list[int]
@click.pass_context
def do(self, task_numbers: List[Tuple]):
    # Convert to ints. Task numbers is a list of tuples. Each tuple contains one digit of the number.
    new_numbers: List[int] = []
    for task_tuple in task_numbers:
        new_number: str = ""
        for task_digit in task_tuple:
            new_number += task_digit
        new_numbers.append(int(new_number))
    # Write back to old value.
    task_numbers = new_numbers
    # Marking something as Done cannot be done with fancy imports
    # So we disable them.
    self.obj['config'].enable_project_folder = False
    try:
        self.obj['task_manager'].import_tasks()
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)
    try:
        completed_tasks = self.obj["task_manager"].mark_as_done(task_numbers)
    except DoneFileNotFound as e:
        handle_done_file_not_found(e, self)
        exit(1)
    # Print the results
    click.echo(self.obj['config'].consoleGood + " Tasks marked as done:")
    for task in completed_tasks:
        click.echo(task)
    commit_exit("do", self.obj["config"])


@root.command('due',
              short_help='List tasks by their due date'
              )
@click.pass_context
def due(self):
    # Disable fancy imports, because they do not have due dates.
    self.obj['config'].enable_project_folder = False
    # Import tasks.
    try:
        self.obj['task_manager'].import_tasks()
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)
    # Print due tree.
    click.echo(self.obj['task_manager'].print_by_due())
    exit()


@root.command('edit',
              short_help='open the todo.txt in $EDITOR (or nano)'
              )
@click.pass_context
def edit(self):
    # Disable fancy imports.
    self.obj['config'].enable_project_folder = False
    program = os.getenv("EDITOR") if os.getenv("EDITOR") else "/usr/bin/nano"
    f = self.obj["config"].todo_file
    click.echo("Editing {} with {}".format(f, program))
    subprocess.run([program, f])
    commit_exit("edit", self.obj["config"])


@root.command('filter',
              short_help='only show tasks containing the search term.'
              )
@click.argument('search_term')
@click.pass_context
def filter_list(self, search_term):
    try:
        self.obj["task_manager"].import_tasks()
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)
    self.obj["task_manager"].filter_by_string(search_term)
    click.echo(self.obj["task_manager"])


@root.command('list', short_help='List tasks')
@click.pass_context
def list_tasks(self):
    try:
        self.obj["task_manager"].import_tasks()
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)
    click.echo(self.obj["task_manager"])


@root.command('list_done', short_help='List tasks which are marked as done')
@click.pass_context
def list_done(self):
    try:
        self.obj["task_manager"].list_done()
    except DoneFileNotFound as e:
        handle_done_file_not_found(e, self)
        exit(1)


@root.command('print_raw', short_help='print todo.txt without any formatting or filtering')
@click.pass_context
def print_raw(self):
    try:
        with open(self.obj["task_manager"].config.todo_file, "r") as f:
            click.echo(f.read())
    except FileNotFoundError as e:
        handle_todo_file_not_found(e, self)


@root.command('priority', short_help='set new priority to task')
@click.argument('task_number', type=int)
@click.argument('new_priority', type=str)
@click.pass_context
def priority(self, task_number, new_priority):
    # Disable fancy imports.
    self.obj['config'].enable_project_folder = False
    # Run task.
    try:
        self.obj["task_manager"].import_tasks()
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)
    self.obj["task_manager"].change_priority(
        priority=(new_priority.upper()), task_number=task_number)


@root.command('project', short_help='print tree by project')
@click.pass_context
def project(self):
    # Import tasks.
    try:
        self.obj["task_manager"].import_tasks()
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)
    # Print due tree.
    click.echo(self.obj['task_manager'].print_project_tree())
    exit()


@root.command('revive', short_help='Revive a task that was accidentally marked as done.')
@click.argument('done_number', type=int)
@click.pass_context
def revive(self, done_number):
    try:
        click.echo(self.obj["task_manager"].revive_task(done_number))
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)
    except DoneFileNotFound as e:
        handle_done_file_not_found(e, self)
        exit(1)
    commit_exit("revive", self.obj["config"])


@root.command('schedule', short_help='hide task until date.',
              help='hide the task until the date given. If new_date is not in ISO format (yyyy-mm-dd), '
                   'then it tries to figure out the date with the `date` program, which is only in linux.'
              )
@click.pass_context
@click.argument('task_number', type=int)
@click.argument('new_date', type=str, nargs=-1)
def schedule(self: click.Context, task_number: int, new_date: Tuple[str]):
    # Disable fancy imports, because they do not have t dates.
    self.obj['config'].enable_project_folder = False
    # Convert
    date = " ".join(new_date)
    date_pattern = re.compile(r"(\d{4})-(\d{2})-(\d{2})")
    if not date_pattern.match(date):
        if self.obj['config'].verbose:
            click.echo(f"{self.obj['config'].consoleGood}Attempt to parse {date} with the `date` program.")
        # Try to use the `date` utility.
        dat = subprocess.run(
            ["date", "-d " + date, "+%F "],
            capture_output=True,
            encoding="utf-8"
        )
        if dat.returncode > 0:
            click.echo(f"{self.obj['config'].consoleError} The date {new_date} could not be parsed.")
            exit(1)
        date = dat.stdout.strip()
    try:
        # Convert to Datetime.
        self.obj["task_manager"].import_tasks()
        self.obj["task_manager"].change_t_date(date, task_number)
    except TodoFileNotFound as e:
        handle_todo_file_not_found(e, self)
        exit(1)

    commit_exit("schedule", self.obj["config"])


@root.group()
@click.pass_context
def stale():
    """
    Use the stale system. NOTE: NOT IMPLEMENTED YET.

    This uses a stale.txt file to store tasks which you want to hide for a long time.
    You will need to manually move them back into todo.txt using this system.
    If you want tasks to disappear for a known time, use the `schedule` command.
    """
    # TODO: Make working again.
    # stale_manager = Taskmanager(
    #    config.consoleGood,
    #    config.consoleWarn,
    #    config.consoleError,
    #    file="stale.txt",
    #    done="todo.txt"
    # )
    # stale_manager.__import_from_todo_txt()


@stale.command('list')
@click.pass_context
def stale_list():
    """
    List tasks in stale.txt NOTE: NOT IMPLEMENTED YET.
    """
    # print(stalemgr)
    pass


@stale.command('restore')
@click.pass_context
def stale_restore(self, task_numbers):
    """
    move task from stale.txt back to todo.txt NOTE: NOT IMPLEMENTED YET.
    :param: task_numbers
    """
    # nums = self.check_ints(task_numbers)
    # taskmgr.revive_task(nums, remove_from_done=True)
    # self.commit_exit(action)


@stale.command('prune', short_help='move task from todo.txt to the stale.txt')
@click.pass_context
def stale_prune(self, task_numbers):
    pass  # TODO: implement.
    # nums = self.check_ints(aux)
    # taskmgr.mark_as_done(nums, change_string=False)
    # self.commit_exit(action)


#  End Region Command Definitions.
#  Setup Click

CONTEXT_SETTINGS: dict = dict(help_option_names=['-h', '--help'])
"""Click context settings. See https://click.palletsprojects.com/en/8.1.x/complex/ for more information."""
cli: click.CommandCollection = click.CommandCollection(
    sources=[root],
    context_settings=CONTEXT_SETTINGS
)
"""Command Collection defining defaults. https://click.palletsprojects.com/en/8.1.x/api/#click.CommandCollection ."""

if __name__ == '__main__':
    cli()



import datetime
import math
import os
import pathlib
import re
import shutil
from tempfile import (  # https://docs.python.org/3/library/tempfile.html
    NamedTemporaryFile,
)
from todotree.Errors.TodoFileNotFound import TodoFileNotFound
from todotree.Errors.DoneFileNotFound import DoneFileNotFound
from todotree.Errors.ProjectFolderError import ProjectFolderError
from todotree.Task import Task
from todotree.Tree import Tree
from todotree.config import Config

# regexes


def task_to_done(tsk):
    """
    Format the task String to a done state, by adding x 2020-02-02 in front of it.
    :param tsk: A task string.
    :return: A done string with a today's time stamp.
    """
    if type(tsk) == Task:  # Flatten to string
        tsk = tsk.raw_string
    done = "x " + str(datetime.date.today()) + " " + tsk
    return done


def task_to_undone(tsk):
    """
    Removes the x 2020-02-02 part so the task can be added to the task file list
    :param tsk: A done string.
    :return: A task string.
    """
    # 13 is the number of chars of "x 2020-02-02 "
    undone = str.strip(tsk[13:])
    return undone


class Taskmanager:
    """Manages the tasks."""

    def __init__(self, configuration: Config = Config()):
        """Initializes a task manager, which manages the tasks."""

        self.task_list: list[Task] = []
        """List Without empty lines"""

        self.config: Config = configuration
        """Configuration settings for the application."""

    # Imports

    def import_tasks(self):
        """
        Import the tasks and projects from files.
        """
        self._import_from_todo_txt()
        if self.config.enable_project_folder:
            self._import_projecttree_folder()

    def _import_from_todo_txt(self):
        """
        Imports the todos from the file to the Taskmanager object.
        """
        try:
            with open(self.config.todo_file, 'r') as f:
                content = f.readlines()
                self.task_list = []
                for i, task in enumerate(content):
                    # Skip empty lines.
                    if task.strip() == "":
                        continue
                    self.task_list.append(Task(i + 1, task.strip()))
        except FileNotFoundError as e:
            raise TodoFileNotFound from e

    def filter_t_date(self):
        """
        Removes Tasks from the Taskmanager where the t:date is later than today.
        """
        start_list = self.task_list.copy()  # Iterable copy.
        for task in start_list:
            if task.t_date is None:
                # It does not have a t date.
                continue
            if task.t_date > datetime.date.today():
                self.task_list.remove(task)

    def filter_block(self):
        """
        Removes Tasks from the Taskmanager where the task is blocked by some other task.
        """
        # Detect all block items.
        block_list = []
        for task in self.task_list:
            block_list.append(task.blocks)
        # Flatten the list of lists to a normal list.
        block_list_items = [item for sublist in block_list for item in sublist]
        # Blocked / Blocked by filtering
        start_list = self.task_list.copy()
        for task in start_list:
            # Check if there is a task which is blocked by another task.
            intersection = list(filter(lambda x: x in block_list_items, task.blocked))
            if intersection:
                self.task_list.remove(task)

    def _import_projecttree_folder(self):
        """
        Adds the directory to the project tree.
        """
        try:
            p = pathlib.Path(self.config.project_tree_folder)
            for i in p.glob("*/"):
                proj = os.path.basename(i)
                # Check if there is a task with that project.
                does_project_task_exist = any(proj in ttt.projects for ttt in self.task_list)
                if not does_project_task_exist:
                    # add gibberish task as placeholder.
                    tmp_task = Task(-1, self.config.emptyProjectString)
                    tmp_task.projects = [proj]
                    self.task_list.append(tmp_task)
        except FileNotFoundError as e:
            raise ProjectFolderError(f"An error occurred while processing the projects folder: {e}")

    #  readonly methods.

    def __str__(self):
        self.task_list.sort(key=lambda x: x.priority)
        self.filter_t_date()
        self.filter_block()
        if self.config.quiet:
            s = ""
        elif self.config.verbose:
            s = self.config.consoleGood + "The todo list is: \n"
        else:
            s = self.config.consoleGood + " Todos\n"

        number_of_digits = int(math.ceil(math.log(len(self.task_list) + 1, 10)))
        for tsk in self.task_list:
            # Number (if applicable).
            s += str(tsk.i).zfill(number_of_digits)
            # Whitespace.
            s += " "
            # Task information self.
            s += str(tsk.stripped_string)
            # new line.
            s += str("\n")
        return s

    def list_done(self):
        """
        List the done.txt with numbers to revive them.
        """
        try:
            with open(self.config.done_file, "r") as f:
                lines = f.readlines()
        except FileNotFoundError as e:
            raise DoneFileNotFound from e
        for i, line in reversed(list(enumerate(lines))):
            print(i, line, end="")

    def filter_by_string(self, filter_string):
        """
        Filters the task list by whether `filter_string` occurred in the task.
        """
        self.task_list = [item for item in self.task_list if filter_string in item.raw_string]

    def print_by_due(self):
        """
        Prints the tree of due dates.
        :return: A string containing the due tree.
        """
        self.filter_t_date()
        self.filter_block()
        return str((Tree(self.task_list, "due_date_band", config=self.config)))

    def print_project_tree(self) -> str:
        """
        print the project tree.
        :return: A string containing the project tree.
        """
        self.filter_t_date()
        self.filter_block()
        return str((Tree(self.task_list, "projects", config=self.config)))

    def print_context_tree(self) -> str:
        """
        print the context tree.
        :return: A string containing the context tree.
        """
        self.filter_t_date()
        self.filter_block()
        return str((Tree(self.task_list, "contexts", config=self.config)))

    #  Write Only Methods

    def write_todo_to_file(self):
        """
        Write the results back to a file.
        """
        # Write to new file.
        self.task_list.sort(key=lambda x: x.i)
        #  Delete=false is needed for windows, I hope that somebodies temp folder won't be clobbered with this..
        try:
            with NamedTemporaryFile("w+t", newline="", delete=False) as temp_file:
                # may strip new lines by using task list.
                for n in self.task_list:
                    temp_file.write(str(n.raw_string))
                    temp_file.write("\n")
                temp_file.flush()
                shutil.copy(temp_file.name, self.config.todo_file)
        except FileNotFoundError as e:
            raise TodoFileNotFound from e

    def change_t_date(self, date: str, task_number: int):
        """
        Adds or changes the t:date
        :param date: The new t:date to add.
        :param task_number: The task number to add the new date to.
        :return: The updated task.
        """
        tsk = self.task_list[task_number - 1]
        tsk.change_t_date(date)
        self.write_todo_to_file()
        return tsk

    def change_priority(self, priority: str, task_number):
        """
        Adds or changes the priority of a task.
        :param priority: The new priority.
        :param task_number: The number of the task.
        :return: The new task.
        """
        tsk = self.task_list[task_number - 1]
        tsk.update_priority(priority)
        self.write_todo_to_file()
        return tsk

    def add_task_to_file(self, tsk: str) -> int:
        """
        Adds a task to a file. Returns the task with its assigned number.
        :param tsk: The task to add.
        :return: The number of the new task.
        """
        try:
            with open(self.config.todo_file, 'a') as f:
                f.write(tsk)
        except FileNotFoundError as e:
            raise TodoFileNotFound from e
        # return the line number
        return len(self.task_list) + 1

    def append_to_task(self, task_number: int, task_string: str) -> str:
        """
        Appends `task_string` to the task defined by `task_number`.
        :param task_string: The string to append
        :param task_number: The number of the existing task.
        :return: The new merged task.
        """
        self.task_list[task_number - 1].append_to_task_string(task_string)
        self.write_todo_to_file()
        return self.task_list[task_number - 1].raw_string

    def mark_as_done(self, line_numbers):
        """
        Marks the task as done, removes it from the task file and adds it to done.txt
        :param line_numbers: The task numbers to mark as done.
        :return: The tasks marked as done.
        """
        # Get and remove completed task
        completed_tasks = []
        line_numbers.sort()
        # Delete the tasks from the end to the start.
        line_numbers.reverse()
        for num in line_numbers:
            completed_tasks.append(self.task_list[num - 1])
            del self.task_list[num - 1]
        # Save the new task list.
        self.write_todo_to_file()
        # Convert completed tasks to done tasks.
        done_tasks = list(map(lambda x: task_to_done(x), completed_tasks))
        # Append the results to done.txt
        try:
            with open(self.config.done_file, "a") as f:
                for completed_task in done_tasks:
                    f.write(completed_task + "\n")
        except FileNotFoundError as e:
            raise DoneFileNotFound from e
        return done_tasks

    def revive_task(self, line_number):
        """
        Gets task from done.txt and adds it to the task file.
        The task does not get removed.
        :param line_number: The done number to revive.
        :return: The revived task.
        """
        # Fetch Task
        try:
            with open(self.config.done_file) as f:
                done_tasks = f.readlines()
        except FileNotFoundError as e:
            raise DoneFileNotFound from e
        tsk = done_tasks[line_number - 1]
        tsk = task_to_undone(tsk)
        # Add task
        self.add_task_to_file(tsk)
        return tsk
